;;; -*- Mode: Scheme -*-

;;;; Module System
;;;; Packages

(define-structure interface
  mapper
  expresser
  (implementations (make-population))
  (name #f)
  )

(define (interface-signature interface identifier)
  ((interface-mapper interface) identifier))

(define (simple-interface exports)
  (make-interface (let ((table (make-identifier-table)))
                    (process-exports exports table)
                    (lambda (identifier)
                      (identifier-table/get table identifier)))
                  (lambda (recur)
                    recur
                    `(EXPORT ,@exports))))

(define (process-exports exports table)
  ((lambda (process-export)
     (for-each process-export exports))
   (lambda (export)
     (if (pair? export)
         (for-each (let ((class (car export)))
                     (lambda (identifier)
                       (identifier-table/put! table identifier class)))
                   (cdr export))
         (identifier-table/put! table export 'VARIABLE)))))

(define (interface-union interfaces)
  (make-interface (lambda (identifier)
                    (let loop ((interfaces interfaces))
                      (and (pair? interfaces)
                           (or (interface-signature (car interfaces)
                                                    identifier)
                               (loop (cdr interfaces))))))
                  (lambda (recur)
                    `(UNION ,@(map recur interfaces)))))

(define (interface/except interface identifiers)
  (make-interface (lambda (identifier)
                    (and (not (memq identifier identifiers))
                         (interface-signature interface identifier)))
                  (lambda (recur)
                    `(EXCEPT ,(recur interface) ,identifiers))))

(define (interface/only interface identifiers)
  (make-interface (lambda (identifier)
                    (and (memq identifier identifiers)
                         (interface-signature interface identifier)))
                  (lambda (recur)
                    `(ONLY ,(recur interface) ,identifiers))))

(define-structure (implementation (constructor %make-implementation))
  interface-thunk
  (%interface #f)
  translation
  package
  (importers (make-population))
  name
  )

(define (make-implementation interface translation package #!optional name)
  (let ((interface (guarantee-interface-thunk interface 'MAKE-IMPLEMENTATION))
        (translation (guarantee-translation translation 'MAKE-IMPLEMENTATION))
        (package (guarantee-package package 'MAKE-IMPLEMENTATION))
        (name (if (default-object? name) #f name)))
    (let ((implementation (%make-implementation interface
                                                translation
                                                package
                                                name)))
      (add-to-population! (package-clients package)
                          implementation)
      implementation)))

(define (initialize-implementation! implementation)
  (set-implementation-%interface!
   implementation
   ((implementation-interface-thunk implementation))))

(define (implementation-lookup implementation identifier)
  (cond ((interface-signature (implementation-interface implementation)
                              identifier)
         => (lambda (signature)
              (let* ((identifier*
                      (translate-identifier
                       (implementation-translation implementation)
                       identifier))
                     (binding
                      (package-lookup (implementation-package implementation)
                                      identifier*)))
                (receive (compatible? signature)
                         (match-signature (binding-signature binding)
                                          signature)
                  (if (not compatible?)
                      (warn:incompatible-signature
                       identifier signature
                       implementation binding))
                  (binding-with-signature binding signature)))))
        (else #f)))

(define (make-implementation name interface translation package)
  (let ((implementation
         (%make-implementation name interface translation package)))
    (add-to-population! (interface-implementations interface)
                        implementation)
    (add-to-population! (package-clients package)
                        implementation)
    implementation))

(define-structure package
  imports-thunk
  (%imports #f)
  tower
  pathname                              ;Merge with files
  files
  (bindings (make-identifier-table))
  environment
  (clients (make-population))           ;Client implementations
  (uid (set! current-package-uid (fix:1+ current-package-uid)))
  )

(define current-package-uid 0)

(define (with-reset-package-uid thunk)
  (fluid-let ((current-package-uid 0))
    (thunk)))

(define (package-imports package)
  (maybe-initialize-package! package)
  (package-%imports package))

(define (initialize-package! package)
  (set-package-%imports! package ((package-imports-thunk package)))
  (for-each (lambda (import)
              (add-to-population! (implementation-importers import)
                                  package))
            (package-%imports package)))

(define (maybe-initialize-package! package)
  (if (not (package-%imports package))
      (initialize-package! package)))

(define (reinitialize-package! package)
  (cond ((package-%imports package)
         => (lambda (imports)
              (for-each (lambda (import)
                          (remove-from-population!
                           (implementation-importers import)
                           package))
                        imports))))
  (initialize-package! package))

(define (package-lookup package identifier)
  (or (identifier-table/get (package-bindings package) identifier)
      (let loop ((imports (package-imports package)))
        (and (pair? imports)
             (or (implementation-lookup (car imports) identifier)
                 (loop (cdr imports)))))))

(define-structure binding
  identifier
  signature
  environment
  )

(define (package-define! package identifier signature)
  (identifier-table/put!
   (package-bindings package)
   identifier
   (make-binding identifier signature package)))

(define (package-define-macro! package identifier item)
  (package-define! package identifier syntax-signature)
  (environment-define-macro (package-environment package)
                            identifier
                            item))
